from qcrash.formatters.base import BaseFormatter
import pytest


def test_format_title():
    title = 'test'
    assert BaseFormatter().format_title(title) == title


def test_format_body():
    with pytest.raises(NotImplementedError):
        assert BaseFormatter().format_body('')



"""
This module containes the email backend.
"""
from .base import BaseBackend
from ..formatters.email import EmailFormatter
from ..qt import QtCore, QtGui


class EmailBackend(BaseBackend):
    """
    This backend sends the crash report via email (using mailto).

    Usage::

        email_backend = qcrash.backends.EmailBackend(
            'your_email@provider.com', 'YourAppName')
        qcrash.install_backend(email_backend)

    """
    def __init__(self, email, app_name, formatter=EmailFormatter()):
        """
        :param email: email address to send the bug report to
        :param app_name: application name, will appear in the object of the
                         mail
        :param formatter: the formatter to use to create the final report.
        """
        super(EmailBackend, self).__init__(
            formatter, "Send email", "Send the report via email",
            QtGui.QIcon.fromTheme('mail-send'), need_review=False)
        self.formatter.app_name = app_name
        self.email = email

    def send_report(self, title, body):
        url = QtCore.QUrl(
            "mailto:%s?subject=%s&body=%s" % (self.email, title, body))
        QtGui.QDesktopServices.openUrl(url)
        return False



from qcrash._dialogs.gh_login import DlgGitHubLogin


def test_github_login_dialog():
    dlg = DlgGitHubLogin(None, '', False)
    assert not dlg.ui.bt_sign_in.isEnabled()
    dlg.ui.le_username.setText('user')
    dlg.ui.le_password.setText('password')
    assert dlg.ui.bt_sign_in.isEnabled()


def test_focus_username(qtbot):
    dlg = DlgGitHubLogin(None, '', False)
    dlg.show()
    assert dlg.ui.le_username.text() == ''
    assert dlg.ui.le_password.text() == ''
    qtbot.waitForWindowShown(dlg)
    dlg.close()


def test_focus_password(qtbot):
    dlg = DlgGitHubLogin(None, 'user', False)
    dlg.show()
    qtbot.waitForWindowShown(dlg)
    assert dlg.ui.le_username.text() == 'user'
    assert dlg.ui.le_password.text() == ''
    dlg.close()



"""
This module contains the base class for implementing a report backend.
"""


class BaseBackend(object):
    """
    Base class for implementing a backend.

    Subclass must define ``button_text``, ``button_tooltip``and ``button_icon``
    and implement ``send_report(title, description)``.

    The report's title and body will be formatted automatically by the
    associated :attr:`formatter`.
    """
    def __init__(self, formatter, button_text, button_tooltip,
                 button_icon=None, need_review=True):
        """
        :param formatter: the associated formatter (see :meth:`set_formatter`)
        :param button_text: Text of the associated button in the report dialog
        :param button_icon: Icon of the associated button in the report dialog
        :param button_tooltip: Tooltip of the associated button in the report
            dialog
        :param need_review: True to show the review dialog before submitting.
            Some backends (such as the email backend) do not need a review
            dialog as the user can already review it before sending the final
            report
        """
        self.formatter = formatter
        self.button_text = button_text
        self.button_tooltip = button_tooltip
        self.button_icon = button_icon
        self.need_review = need_review

    def qsettings(self):
        """
        Gets the qsettings instance that you can use to store various settings
        such as the user credentials (you should use the `keyring` module if
        you want to store user's password).
        """
        from qcrash.api import _qsettings
        return _qsettings

    def set_formatter(self, formatter):
        """
        Sets the formatter associated with the backend.

        The formatter will automatically get called to format the report title
        and body before ``send_report`` is being called.
        """
        self.formatter = formatter

    def send_report(self, title, body):
        """
        Sends the actual bug report.

        :param title: title of the report, already formatted.
        :param body: body of the reporit, already formtatted.

        :returns: Whether the dialog should be closed.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/colin/dev/qcrash/forms/dlg_report_bug.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from qcrash.qt import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(544, 272)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.ExpandingFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.label = QtWidgets.QLabel(Dialog)
        self.label.setObjectName("label")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label)
        self.lineEditTitle = QtWidgets.QLineEdit(Dialog)
        self.lineEditTitle.setObjectName("lineEditTitle")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.lineEditTitle)
        self.label_2 = QtWidgets.QLabel(Dialog)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.plainTextEditDesc = QtWidgets.QPlainTextEdit(Dialog)
        self.plainTextEditDesc.setObjectName("plainTextEditDesc")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.plainTextEditDesc)
        self.cb_include_sys_info = QtWidgets.QCheckBox(Dialog)
        self.cb_include_sys_info.setObjectName("cb_include_sys_info")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.cb_include_sys_info)
        self.cb_include_application_log = QtWidgets.QCheckBox(Dialog)
        self.cb_include_application_log.setObjectName("cb_include_application_log")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.cb_include_application_log)
        self.verticalLayout.addLayout(self.formLayout)
        self.layout_buttons = QtWidgets.QHBoxLayout()
        self.layout_buttons.setContentsMargins(-1, 0, -1, -1)
        self.layout_buttons.setObjectName("layout_buttons")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.layout_buttons.addItem(spacerItem)
        self.verticalLayout.addLayout(self.layout_buttons)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Report an issue..."))
        self.label.setText(_translate("Dialog", "Title:"))
        self.lineEditTitle.setToolTip(_translate("Dialog", "Title of the issue"))
        self.label_2.setText(_translate("Dialog", "Description:"))
        self.plainTextEditDesc.setToolTip(_translate("Dialog", "Description of the issue (mandatory)"))
        self.cb_include_sys_info.setToolTip(_translate("Dialog", "Include system information"))
        self.cb_include_sys_info.setText(_translate("Dialog", "Include system &information"))
        self.cb_include_application_log.setToolTip(_translate("Dialog", "Include application log"))
        self.cb_include_application_log.setText(_translate("Dialog", "Include application &log"))

from . import qcrash_rc


"""
This package contains the backends used to actually send the crash report.

QCrash provides 2 backends:

- email: let the user report the crash/issue via an email
- github: let the user report the crash/issue on you github issue tracker

To use a backend, use :meth:`qcrash.api.install_backend`.

To add your own, just subclass :class:`qcrash.backends.BaseBackend` and
implement `send_report`.
"""
from .base import BaseBackend
from .email import EmailBackend
from .github import GithubBackend


__all__ = [
    'BaseBackend',
    'EmailBackend',
    'GithubBackend'
]



from qcrash._dialogs.review import DlgReview


def test_review(qtbot):
    dlg = DlgReview('some content', None)
    assert dlg.ui.plainTextEdit.toPlainText() == 'some content'
    qtbot.keyPress(dlg.ui.plainTextEdit, 'A')
    assert dlg.ui.plainTextEdit.toPlainText() == 'Asome content'



"""
This module contains the top level API functions.
"""
from . import _hooks
from .qt import QtCore

from . import backends


def install_backend(*args):
    """
    Install one or more backends.

    Usage::

        qcrash.install_backend(backend1)
        qcrash.install_backend(backend2, backend3)

    :param args: the backends to install. Each backend must be a subclass
        of :class:`qcrash.backends.BaseBackend` (e.g.::
        :class:`qcrash.backends.EmailBackend` or
        :class:`qcrash.backends.GithubBackend`)
    """
    global _backends
    for b in args:
        _backends.append(b)


def get_backends():
    """
    Gets the list of installed backends.
    """
    return _backends


def install_except_hook(except_hook=_hooks.except_hook):
    """
    Install an except hook that will show the crash report dialog when an
    unhandled exception has occured.

    :param except_hook: except_hook function that will be called on the main
        thread whenever an unhandled exception occured. The function takes
        two parameters: the exception object and the traceback string.
    """
    if not _backends:
        raise ValueError('no backends found, you must at least install one '
                         'backend before calling this function')
    global _except_hook
    _except_hook = _hooks.QtExceptHook(except_hook)


def set_qsettings(qsettings):
    """
    Sets the qsettings used by the backends to cache some information such
    as the user credentials. If no custom qsettings is defined, qcrash will
    use its own settings (QSettings('QCrash'))

    :param qsettings: QtCore.QSettings instance
    """
    global _qsettings
    _qsettings = qsettings


def show_report_dialog(window_title='Report an issue...',
                       window_icon=None, traceback=None, issue_title='',
                       issue_description='', parent=None,
                       modal=None, include_log=True, include_sys_info=True):
    """
    Show the issue report dialog manually.

    :param window_title: Title of dialog window
    :param window_icon: the icon to use for the dialog window
    :param traceback: optional traceback string to include in the report.
    :param issue_title: optional issue title
    :param issue_description: optional issue description
    :param parent: parent widget
    :param include_log: Initial state of the include log check box
    :param include_sys_info: Initial state of the include system info check box
    """
    if not _backends:
        raise ValueError('no backends found, you must at least install one '
                         'backend before calling this function')
    from ._dialogs.report import DlgReport
    dlg = DlgReport(_backends, window_title=window_title,
                    window_icon=window_icon, traceback=traceback,
                    issue_title=issue_title,
                    issue_description=issue_description, parent=parent,
                    include_log=include_log, include_sys_info=include_sys_info)
    if modal:
        dlg.show()
        return dlg
    else:
        dlg.exec_()


def _return_empty_string():
    return ''


#: Reference to the function to use to collect system information. Client code
#: should redefine it.
get_system_information = _return_empty_string

#: Reference to the function to use to collect the application's log.
#: Client code should redefine it.
get_application_log = _return_empty_string


_backends = []
_qsettings = QtCore.QSettings('QCrash')


__all__ = [
    'backends',
    'install_backend',
    'get_backends',
    'install_except_hook',
    'set_qsettings',
    'show_report_dialog',
    'get_application_log',
    'get_system_information',
]



from qcrash.formatters.markdown import MardownFormatter


def test_format_body():
    description = 'A description'
    traceback = 'A traceback'
    sys_info = '''OS: Linux
Python: 3.4.1
Qt: 5.5.1'''
    log = 'blabla'
    expected = '''### Description

%s

### Traceback

```
%s
```

### System information

- OS: Linux
- Python: 3.4.1
- Qt: 5.5.1

### Application log

```
%s
```

''' % (description, traceback, log)

    assert MardownFormatter().format_body(
        description, sys_info, log, traceback) == expected


def test_format_body_no_log_no_sys_info():
    description = 'A description'
    traceback = 'A traceback'
    sys_info = None
    log = None
    expected = '''### Description

%s

### Traceback

```
%s
```

''' % (description, traceback)

    assert MardownFormatter().format_body(
        description, sys_info, log, traceback) == expected



import logging
import sys

_logger = logging.getLogger(__name__)

try:
    from PyQt5 import QtWidgets, QtGui, QtCore
except (ImportError, RuntimeError):
    _logger.warning('failed to import PyQt5, going to try PyQt4')
    try:
        from PyQt4 import QtGui, QtGui as QtWidgets, QtCore
    except (ImportError, RuntimeError):
        _logger.warning('failed to import PyQt4, going to try PySide')
        try:
            from PySide import QtGui, QtCore, QtGui as QtWidgets
        except (ImportError, RuntimeError):
            _logger.warning('failed to import PySide')
            _logger.critical('No Qt bindings found, aborting...')
            try:
                from unittest.mock import MagicMock
                QtCore = MagicMock()
                QtGui = MagicMock()
                QtWidgets = MagicMock()
            except ImportError:
                sys.exit(1)

__all__ = ['QtCore', 'QtGui', 'QtWidgets']



from qcrash._dialogs.report import DlgReport

import pytest


@pytest.mark.parametrize('include_log', [True, False])
def test_include_log_param(include_log):
    dlg = DlgReport([], include_log=include_log)
    assert dlg.ui.cb_include_application_log.isChecked() == include_log


@pytest.mark.parametrize('include_sys_info', [True, False])
def test_include_sys_info_param(include_sys_info):
    dlg = DlgReport([], include_sys_info=include_sys_info)
    assert dlg.ui.cb_include_sys_info.isChecked() == include_sys_info



"""
This package two hooks functions:

- a hook for pyqt distutis to replace PyQt5 imports by our own (qcrash.qt).
- a sys.excepthook is installed by :meth:`qcrash.api.install_except_hook`
"""
import logging
import sys
import traceback

from .qt import QtCore, QtWidgets


def _logger():
    return logging.getLogger(__name__)


try:
    Signal = QtCore.pyqtSignal
except AttributeError:
    Signal = QtCore.Signal


def fix_qt_imports(path):
    with open(path, 'r') as f_script:
        lines = f_script.read().splitlines()
    new_lines = []
    for l in lines:
        if l.startswith("import "):
            l = "from . " + l
        if "from PyQt5 import" in l:
            l = l.replace("from PyQt5 import", "from qcrash.qt import")
        new_lines.append(l)
    with open(path, 'w') as f_script:
        f_script.write("\n".join(new_lines))


def except_hook(exc, tb):
    from qcrash.api import show_report_dialog
    title = '[Unhandled exception]  %s: %s' % (
        exc.__class__.__name__, str(exc))
    msg_box = QtWidgets.QMessageBox()
    msg_box.setWindowTitle('Unhandled exception')
    msg_box.setText('An unhandled exception has occured...')
    msg_box.setInformativeText(
        'Would you like to report the bug to the developers?')
    msg_box.setIcon(msg_box.Critical)
    msg_box.setDetailedText(tb)
    msg_box.setStandardButtons(QtWidgets.QMessageBox.Ok |
                               QtWidgets.QMessageBox.Cancel)
    msg_box.button(msg_box.Ok).setText('Report')
    msg_box.button(msg_box.Cancel).setText('Close')
    if msg_box.exec_() == msg_box.Ok:
        show_report_dialog(window_title='Report unhandled exception',
                           issue_title=title, traceback=tb)


class QtExceptHook(QtCore.QObject):
    _report_exception_requested = Signal(object, object)

    def __init__(self, except_hook, *args, **kwargs):
        super(QtExceptHook, self).__init__(*args, **kwargs)
        sys.excepthook = self._except_hook
        self._report_exception_requested.connect(except_hook)

    def _except_hook(self, exc_type, exc_val, tb):
        tb = '\n'.join([''.join(traceback.format_tb(tb)),
                        '{0}: {1}'.format(exc_type.__name__, exc_val)])
        _logger().critical('unhandled exception:\n%s', tb)
        # exception might come from another thread, use a signal
        # so that we can be sure we will show the bug report dialog from
        # the main gui thread.
        self._report_exception_requested.emit(exc_val, tb)



from qcrash.backends.base import BaseBackend
from qcrash.formatters.email import EmailFormatter

import pytest


def test_qsettings():
    b = BaseBackend(None, '', '', None)
    assert b.qsettings() is not None


def test_set_formatter():
    b = BaseBackend(None, '', '', None)
    assert b.formatter is None
    b.set_formatter(EmailFormatter("test"))
    assert isinstance(b.formatter, EmailFormatter)


def test_send_report():
    b = BaseBackend(None, '', '', None)
    with pytest.raises(NotImplementedError):
        b.send_report('', '')



import sys
import pytest

from qcrash import api
from qcrash.qt import QtCore
from qcrash._hooks import QtExceptHook


def test_set_qsettings():
    assert api._qsettings.organizationName() == 'QCrash'
    api.set_qsettings(QtCore.QSettings('TestQCrash'))
    assert api._qsettings.organizationName() == 'TestQCrash'


def test_value_errors():
    api._backends[:] = []
    with pytest.raises(ValueError):
        api.install_except_hook()
    with pytest.raises(ValueError):
        api.show_report_dialog()


def test_install_backend():
    api._backends[:] = []
    assert len(api.get_backends()) == 0
    b1 = api.backends.GithubBackend('ColinDuquesnoy', 'TestQCrash')
    b2 = api.backends.EmailBackend('colin.duquesnoy@gmail.com', 'TestQCrash')
    api.install_backend(b1, b2)
    assert len(api.get_backends()) == 2
    b1 = api.backends.GithubBackend('ColinDuquesnoy', 'TestQCrash')
    b2 = api.backends.EmailBackend('colin.duquesnoy@gmail.com', 'TestQCrash')
    api.install_backend(b1)
    assert len(api.get_backends()) == 3
    api.install_backend(b2)
    assert len(api.get_backends()) == 4


def test_install_except_hook():
    api.install_except_hook()
    assert isinstance(api._except_hook, QtExceptHook)
    assert sys.excepthook == api._except_hook._except_hook


def test_show_report_dialog(qtbot):
    api._backends[:] = []
    assert len(api.get_backends()) == 0
    b1 = api.backends.GithubBackend('ColinDuquesnoy', 'TestQCrash')
    b2 = api.backends.EmailBackend('colin.duquesnoy@gmail.com', 'TestQCrash')
    api.install_backend(b1, b2)
    assert len(api.get_backends()) == 2

    dlg = api.show_report_dialog(modal=True)

    assert len(dlg.buttons) == 2

    # force email backend to return True to close the dialog
    b2.old_send_report = b2.send_report

    qtbot.wait(1000)

    def send_report(*args, **kwargs):
        b2.old_send_report(*args, **kwargs)
        return True

    b2.send_report = send_report
    dlg.buttons[1].clicked.emit(True)

    dlg.reject()


def test_return_empty_string():
    assert api._return_empty_string() == ''



"""
This module contains the Html formatter used by the email backend.
"""
from .base import BaseFormatter


BODY_ITEM_TEMPLATE = '''%(name)s
%(delim)s

%(value)s


'''

NB_LINES_MAX = 50


class EmailFormatter(BaseFormatter):
    """
    Formats the crash report for use in an email (text/plain)
    """
    def __init__(self, app_name=None):
        """
        :param app_name: Name of the application. If set the email subject will
             starts with [app_name]
        """
        self.app_name = app_name

    def format_title(self, title):
        """
        Formats title (add ``[app_name]`` if app_name is not None).
        """
        if self.app_name:
            return '[%s] %s' % (self.app_name, title)
        return title

    def format_body(self, description, sys_info=None, log=None,
                    traceback=None):
        """
        Formats the body in plain text. (add a series of '-' under each section
            title).

        :param description: Description of the issue, written by the user.
        :param sys_info: Optional system information string
        :param log: Optional application log
        :param traceback: Optional traceback.
        """
        name = 'Description'
        delim = '-' * 40
        body = BODY_ITEM_TEMPLATE % {
            'name': name, 'value': description, 'delim': delim
        }
        if traceback:
            name = 'Traceback'
            traceback = '\n'.join(traceback.splitlines()[-NB_LINES_MAX:])
            body += BODY_ITEM_TEMPLATE % {
                'name': name, 'value': traceback, 'delim': delim
            }
        if sys_info:
            name = 'System information'
            body += BODY_ITEM_TEMPLATE % {
                'name': name, 'value': sys_info, 'delim': delim
            }
        if log:
            name = 'Application log'
            log = '\n'.join(log.splitlines()[-NB_LINES_MAX:])
            body += BODY_ITEM_TEMPLATE % {
                'name': name, 'value': log, 'delim': delim
            }
        return body



"""
This module contains the review dialog.
"""
from qcrash.qt import QtWidgets
from qcrash._forms import dlg_review_ui


class DlgReview(QtWidgets.QDialog):
    """
    Dialog for reviewing the final report.
    """
    def __init__(self, content, parent):
        """
        :param content: content of the final report, before review
        :param parent: parent widget
        """
        super(DlgReview, self).__init__(parent)
        self.ui = dlg_review_ui.Ui_Dialog()
        self.ui.setupUi(self)
        self.ui.plainTextEdit.setPlainText(content)

    @classmethod
    def review(cls, content, parent):  # pragma: no cover
        """
        Reviews the final bug report.

        :param content: content of the final report, before review
        :param parent: parent widget

        :returns: the reviewed report content or None if the review was
                  canceled.
        """
        dlg = DlgReview(content, parent)
        if dlg.exec_():
            return dlg.ui.plainTextEdit.toPlainText()
        return None



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/colin/dev/QCrash/forms/dlg_github_login.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from qcrash.qt import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(400, 222)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.lbl_html = QtWidgets.QLabel(Dialog)
        self.lbl_html.setObjectName("lbl_html")
        self.verticalLayout.addWidget(self.lbl_html)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label_2 = QtWidgets.QLabel(Dialog)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout.addWidget(self.label_2)
        self.le_username = QtWidgets.QLineEdit(Dialog)
        self.le_username.setObjectName("le_username")
        self.horizontalLayout.addWidget(self.le_username)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setContentsMargins(-1, 0, -1, -1)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_3 = QtWidgets.QLabel(Dialog)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_2.addWidget(self.label_3)
        self.le_password = QtWidgets.QLineEdit(Dialog)
        self.le_password.setEchoMode(QtWidgets.QLineEdit.Password)
        self.le_password.setObjectName("le_password")
        self.horizontalLayout_2.addWidget(self.le_password)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.cb_remember = QtWidgets.QCheckBox(Dialog)
        self.cb_remember.setObjectName("cb_remember")
        self.verticalLayout.addWidget(self.cb_remember)
        self.bt_sign_in = QtWidgets.QPushButton(Dialog)
        self.bt_sign_in.setObjectName("bt_sign_in")
        self.verticalLayout.addWidget(self.bt_sign_in)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Sign in to github"))
        self.lbl_html.setText(_translate("Dialog", "<html><head/><body><p align=\"center\"><img src=\":/rc/GitHub-Mark.png\"/></p><p align=\"center\">Sign in to GitHub</p></body></html>"))
        self.label_2.setText(_translate("Dialog", "Username:"))
        self.label_3.setText(_translate("Dialog", "Password: "))
        self.cb_remember.setText(_translate("Dialog", "Remember me"))
        self.bt_sign_in.setText(_translate("Dialog", "Sign in"))

from . import qcrash_rc


"""
This module contains the github backend.
"""
import logging
import webbrowser

import keyring

from .base import BaseBackend
from ..formatters.markdown import MardownFormatter
from ..qt import QtGui, QtWidgets
from .._dialogs.gh_login import DlgGitHubLogin
from .._extlibs import github


GH_MARK_NORMAL = ':/rc/GitHub-Mark.png'
GH_MARK_LIGHT = ':/rc/GitHub-Mark-Light.png'


def _logger():
    return logging.getLogger(__name__)


class GithubBackend(BaseBackend):
    """
    This backend sends the crash report on a github issue tracker::

        https://github.com/gh_owner/gh_repo

    Usage::

        github_backend = qcrash.backends.GithubBackend(
            'ColinDuquesnoy', 'QCrash')
        qcrash.install_backend(github_backend)
    """
    def __init__(self, gh_owner, gh_repo, formatter=MardownFormatter()):
        """
        :param gh_owner: Name of the owner of the github repository.
        :param gh_repo: Name of the repository on github.
        """
        super(GithubBackend, self).__init__(
            formatter, "Submit on github",
            "Submit the issue on our issue tracker on github", None)
        icon = GH_MARK_NORMAL
        if QtWidgets.qApp.palette().base().color().lightness() < 128:
            icon = GH_MARK_LIGHT
        self.button_icon = QtGui.QIcon(icon)
        self.gh_owner = gh_owner
        self.gh_repo = gh_repo
        self._show_msgbox = True  # False when running the test suite

    def send_report(self, title, body):
        _logger().debug('sending bug report on github\ntitle=%s\nbody=%s',
                        title, body)
        username, password, remember = self.get_user_credentials()
        if not username or not password:
            return False
        _logger().debug('got user credentials')
        try:
            gh = github.GitHub(username=username, password=password)
            repo = gh.repos(self.gh_owner)(self.gh_repo)
            ret = repo.issues.post(title=title, body=body)
        except github.ApiError as e:
            _logger().warn('failed to send bug report on github. response=%r' %
                           e.response)
            # invalid credentials
            if e.response.code == 401:
                self.qsettings().setValue('github/remember_credentials', 0)
                if self._show_msgbox:
                    QtWidgets.QMessageBox.warning(
                      QtWidgets.qApp.activeWindow(), 'Invalid credentials',
                      'Failed to create github issue, invalid credentials...')
            else:
                # other issue
                if self._show_msgbox:
                    QtWidgets.QMessageBox.warning(
                        QtWidgets.qApp.activeWindow(),
                        'Failed to create issue',
                        'Failed to create github issue. Error %d' %
                        e.response.code)
            return False
        else:
            issue_nbr = ret['number']
            if self._show_msgbox:
                ret = QtWidgets.QMessageBox.question(
                    QtWidgets.qApp.activeWindow(), 'Issue created on github',
                    'Issue successfully created. Would you like to open the '
                    'ticket in your web browser?')
            if ret in [QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.Ok]:
                webbrowser.open(
                    'https://github.com/%s/%s/issues/%d' % (
                        self.gh_owner, self.gh_repo, issue_nbr))
            return True

    def _get_credentials_from_qsettings(self):
        remember = self.qsettings().value('github/remember_credentials', "0")
        username = self.qsettings().value('github/username', "")
        try:
            # PyQt5 or PyQt4 api v2
            remember = bool(int(remember))
        except TypeError:  # pragma: no cover
            # pyside returns QVariants
            remember, _ok = remember.toInt()
            username = username.toString()
        return username, bool(remember)

    def _store_credentials(self, username, password, remember):
        self.qsettings().setValue('github/username', username)
        try:
            keyring.set_password('github', username, password)
        except RuntimeError:  # pragma: no cover
            _logger().warn('failed to save password in keyring, you '
                           'will be prompted for your credentials '
                           'next time you want to report an issue')
            remember = False
        self.qsettings().setValue(
            'github/remember_credentials', int(remember))

    def get_user_credentials(self):  # pragma: no cover
        # reason: hard to test methods that shows modal dialogs
        username, remember = self._get_credentials_from_qsettings()

        if remember and username:
            # get password from keyring
            try:
                password = keyring.get_password('github', username)
            except RuntimeError:
                # no safe keyring backend
                _logger().warn('failed to retrieve password from keyring...')
            else:
                return username, password, remember

        # ask for credentials
        username, password, remember = DlgGitHubLogin.login(
            QtWidgets.qApp.activeWindow(), username, remember)

        if remember:
            self._store_credentials(username, password, remember)

        return username, password, remember



"""
A PyQt/PySide framework for reporting application crash (unhandled exception)
and let the user report an issue/feature request.
"""
from .base import BaseFormatter


BODY_ITEM_TEMPLATE = '''### %(name)s

%(value)s

'''

NB_LINES_MAX = 50


class MardownFormatter(BaseFormatter):
    """
    Formats the issue report using Markdown.
    """
    def format_body(self, description, sys_info=None, log=None,
                    traceback=None):
        """
        Formats the body using markdown.

        :param description: Description of the issue, written by the user.
        :param sys_info: Optional system information string
        :param log: Optional application log
        :param traceback: Optional traceback.
        """
        body = BODY_ITEM_TEMPLATE % {
            'name': 'Description', 'value': description
        }
        if traceback:
            traceback = '\n'.join(traceback.splitlines()[-NB_LINES_MAX:])
            body += BODY_ITEM_TEMPLATE % {
                'name': 'Traceback', 'value': '```\n%s\n```' % traceback
            }
        if sys_info:
            sys_info = '- %s' % '\n- '.join(sys_info.splitlines())
            body += BODY_ITEM_TEMPLATE % {
                'name': 'System information', 'value': sys_info
            }
        if log:
            log = '\n'.join(log.splitlines()[-NB_LINES_MAX:])
            body += BODY_ITEM_TEMPLATE % {
                'name': 'Application log', 'value': '```\n%s\n```' % log
            }
        return body



class BaseFormatter(object):
    """
    Base class for implementing a custom formatter.

    Just implement :meth:`format_body` and :meth:`format_title` functions and
    set your formatter on the backends you created.
    """
    def format_title(self, title):
        """
        Formats the issue title. By default this method does nothing.

        An email formatter might want to append the application to name to
        the object field...
        """
        return title

    def format_body(self, description, sys_info=None, log=None,
                    traceback=None):
        """
        Not implemented.

        :param description: Description of the issue, written by the user.
        :param sys_info: Optional system information string
        :param log: Optional application log
        :param traceback: Optional traceback.
        """
        raise NotImplementedError



import logging

from qcrash._forms import dlg_report_bug_ui
from qcrash.qt import QtGui, QtWidgets
from qcrash._dialogs.review import DlgReview


_logger = logging.getLogger(__name__)


LOG_LIMIT = 100  # keep only the last 100 lines


class DlgReport(QtWidgets.QDialog):
    """
    A qt dialog for reporting an issue on github or via email.

    The dialog let the user choose whether he wants to report an issue or
    propose an enhancement. Depending on its choice the issue title will be
    prefixed by "[Bug]" or "Enhancement". The user must write a description
    of the issue. User can choose to send the report via email or on github
    (using the user's credentials).

    The client code can specify some additional information to be included in
    the report (but hidden in thr dialog).

        - sys_infos: information about the system
        - traceback: the traceback of an unhandled exception
        - log: the application log (will be truncated).

    """

    def __init__(self, backends, window_title='Report an issue...',
                 window_icon=None, traceback=None, issue_title='',
                 issue_description='', include_log=True, include_sys_info=True,
                 **kwargs):
        """
        """
        super(DlgReport, self).__init__(**kwargs)
        self._traceback = traceback
        self.ui = dlg_report_bug_ui.Ui_Dialog()
        self.ui.setupUi(self)
        self.ui.cb_include_sys_info.setChecked(include_sys_info)
        self.ui.cb_include_application_log.setChecked(include_log)
        self.setWindowTitle(window_title)
        self.setWindowIcon(QtGui.QIcon.fromTheme('tools-report-bug')
                           if window_icon is None else window_icon)
        self.ui.lineEditTitle.setText(issue_title)
        self.ui.plainTextEditDesc.setPlainText(issue_description)

        self.ui.lineEditTitle.textChanged.connect(self._enable_buttons)
        self.ui.plainTextEditDesc.textChanged.connect(self._enable_buttons)

        self.buttons = []
        for backend in backends:
            bt = QtWidgets.QPushButton()
            bt.setText(backend.button_text)
            if backend.button_icon:
                bt.setIcon(backend.button_icon)
            bt.backend = backend
            bt.clicked.connect(self._on_button_clicked)
            self.ui.layout_buttons.addWidget(bt)
            self.buttons.append(bt)
        self._enable_buttons()

    def _enable_buttons(self, *_):
        title = str(self.ui.lineEditTitle.text()).strip()
        desc = str(self.ui.plainTextEditDesc.toPlainText()).strip()
        enable = title != '' and desc != ''
        for bt in self.buttons:
            bt.setEnabled(enable)

    def _on_button_clicked(self):
        from qcrash import api
        bt = self.sender()
        description = self.ui.plainTextEditDesc.toPlainText()
        backend = bt.backend
        title = backend.formatter.format_title(
            str(self.ui.lineEditTitle.text()))

        sys_info = None
        if self.ui.cb_include_sys_info.isChecked():
            sys_info = api.get_system_information()

        log = None
        if self.ui.cb_include_application_log.isChecked():
            log = api.get_application_log()

        body = backend.formatter.format_body(
            str(description), sys_info, log, self._traceback)

        if backend.need_review:  # pragma: no cover
            body = DlgReview.review(body, self)
            if body is None:
                return  # user cancelled the review dialog

        if backend.send_report(title, body):
            self.accept()



from qcrash.backends.email import EmailBackend


EMAIL = 'your.email@provider.com'


def get_backend():
    return EmailBackend(EMAIL, 'TestQCrash')


def test_send_report():
    b = get_backend()
    ret = b.send_report('A title', 'A body')
    assert ret is False  # means report dialog must not be closed



"""
This package contains the formatters used by the different backends:

- html_formatter: create a bug report using HTML
- md_formatter: create a bug report using Markdown

Implement :class:`qcrash.formatters.BaseFormatter` to add your own formatter
and set it the installed backends using
:meth:`qcrash.backends.BaseBackend.set_formatter`.

"""



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/colin/dev/qcrash/forms/dlg_review.ui'
#
# Created by: PyQt5 UI code generator 5.5.1
#
# WARNING! All changes made in this file will be lost!

from qcrash.qt import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(622, 495)
        icon = QtGui.QIcon.fromTheme("document-edit")
        Dialog.setWindowIcon(icon)
        Dialog.setToolTip("")
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(Dialog)
        self.label.setStyleSheet("background-color:palette(highlight) ;\n"
"color: palette(highlighted-text);\n"
"padding: 10px;\n"
"border-radius:3px;")
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.plainTextEdit = QtWidgets.QPlainTextEdit(Dialog)
        self.plainTextEdit.setObjectName("plainTextEdit")
        self.verticalLayout.addWidget(self.plainTextEdit)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Review"))
        self.label.setText(_translate("Dialog", "<html><head/><body><p align=\"center\">Review the final report</p></body></html>"))



from qcrash.formatters.email import EmailFormatter


def test_format_title():
    title = 'test'
    appname = 'TestQCrash'
    expected = '[%s] %s' % (appname, title)
    assert EmailFormatter(app_name=appname).format_title(title) == expected
    assert EmailFormatter().format_title(title) == title


def test_format_body():
    appname = 'TestQCrash'
    description = 'A description'
    traceback = 'A traceback'
    sys_info = '''OS: Linux
Python: 3.4.1
Qt: 5.5.1'''
    log = 'blabla'
    expected = '''Description
----------------------------------------

%s


Traceback
----------------------------------------

%s


System information
----------------------------------------

%s


Application log
----------------------------------------

%s


''' % (description, traceback, sys_info, log)

    assert EmailFormatter(app_name=appname).format_body(
        description, sys_info, log, traceback) == expected


def test_format_body_no_log_no_sys_info():
    appname = 'TestQCrash'
    description = 'A description'
    traceback = 'A traceback'
    sys_info = None
    log = None
    expected = '''Description
----------------------------------------

%s


Traceback
----------------------------------------

%s


''' % (description, traceback)

    assert EmailFormatter(app_name=appname).format_body(
        description, sys_info, log, traceback) == expected



#!/usr/bin/env python
"""
Setup script for QCrash
"""
import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
from qcrash import __version__

try:
    from pyqt_distutils.build_ui import build_ui
    cmdclass = {'build_ui': build_ui}
except ImportError:
    build_ui = None
    cmdclass = {}


class PyTest(TestCommand, object):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        super(PyTest, self).initialize_options()
        self.pytest_args = []

    def finalize_options(self):
        super(PyTest, self).finalize_options()

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


cmdclass['test'] = PyTest


# Get long description
with open('README.rst', 'r') as readme:
    long_desc = readme.read()

setup(
    name='qcrash',
    version=__version__,
    keywords=['Github', 'PyQt4', 'PyQt5', 'PySide', 'Issue', 'Report', 'Crash',
              'Tool'],
    url='https://github.com/ColinDuquesnoy/qcrash',
    license='MIT',
    author='Colin Duquesnoy',
    author_email='colin.duquesnoy@gmail.com',
    description='A crash report framework for PyQt/PySide applications',
    long_description=long_desc,
    packages=find_packages(),
    cmdclass=cmdclass,
    install_requires=['keyring'],
    tests_require=['pytest', 'pytest-qt', 'pytest-cov', 'pytest-flake8'],
    entry_points={
        'pyqt_distutils_hooks': [
            'fix_qt_imports = qcrash._hooks:fix_qt_imports']
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: X11 Applications :: Qt',
        'Environment :: Win32 (MS Windows)',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later '
        '(GPLv3+)',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Bug Tracking',
        'Topic :: Software Development :: User Interfaces',
        'Topic :: Software Development :: Widget Sets']
)



from qcrash.qt import QtWidgets

from qcrash._forms import dlg_github_login_ui


GH_MARK_NORMAL = ':/rc/GitHub-Mark.png'
GH_MARK_LIGHT = ':/rc/GitHub-Mark-Light.png'


class DlgGitHubLogin(QtWidgets.QDialog):
    HTML = '<html><head/><body><p align="center"><img src="%s"/></p>' \
        '<p align="center">Sign in to GitHub</p></body></html>'

    def __init__(self, parent, username, remember):
        super(DlgGitHubLogin, self).__init__(parent)
        self.ui = dlg_github_login_ui.Ui_Dialog()

        self.ui.setupUi(self)

        mark = GH_MARK_NORMAL
        if self.palette().base().color().lightness() < 128:
            mark = GH_MARK_LIGHT
        html = self.HTML % mark
        self.ui.lbl_html.setText(html)
        self.ui.bt_sign_in.clicked.connect(self.accept)
        self.ui.le_username.textChanged.connect(self.update_btn_state)
        self.ui.le_password.textChanged.connect(self.update_btn_state)
        self.ui.bt_sign_in.setDisabled(True)
        self.ui.le_username.setText(username)
        self.ui.cb_remember.setChecked(remember)
        if username:
            self.ui.le_password.setFocus()
        else:
            self.ui.le_username.setFocus()

    def update_btn_state(self):
        enable = str(self.ui.le_username.text()).strip() != ''
        enable &= str(self.ui.le_password.text()).strip() != ''
        self.ui.bt_sign_in.setEnabled(enable)

    @classmethod
    def login(cls, parent, username, remember):  # pragma: no cover
        dlg = DlgGitHubLogin(parent, username, remember)
        if dlg.exec_() == dlg.Accepted:
            return dlg.ui.le_username.text(), dlg.ui.le_password.text(), \
                dlg.ui.cb_remember.isChecked()
        return None, None, None



from qcrash import api, qt

USERNAME = 'QCrash-Tester'
PASSWORD = 'TestQCrash1234'
GH_OWNER = 'ColinDuquesnoy'
GH_REPO = 'QCrash-Test'


def get_backend():
    b = api.backends.GithubBackend(GH_OWNER, GH_REPO)
    b._show_msgbox = False
    return b


def get_backend_bad_repo():
    b = api.backends.GithubBackend(GH_OWNER, GH_REPO + '1234')
    b._show_msgbox = False
    return b


def get_wrong_user_credentials():
    """
    Monkeypatch GithubBackend.get_user_credentials to force the case where
    invalid credentias were provided
    """
    return 'invalid', 'invalid', False


def get_empty_user_credentials():
    """
    Monkeypatch GithubBackend.get_user_credentials to force the case where
    invalid credentias were provided
    """
    return '', '', False


def get_fake_user_credentials():
    """
    Monkeypatch GithubBackend.get_user_credentials to force the case where
    invalid credentias were provided
    """
    return USERNAME, PASSWORD, False


def test_invalid_credentials():
    b = get_backend()
    b.get_user_credentials = get_wrong_user_credentials
    ret = b.send_report('Wrong credentials', 'Wrong credentials')
    assert ret is False


def test_empty_credentials():
    b = get_backend()
    b.get_user_credentials = get_empty_user_credentials
    ret = b.send_report('Empty credentials', 'Wrong credentials')
    assert ret is False


def test_fake_credentials():
    b = get_backend()
    b.get_user_credentials = get_fake_user_credentials
    ret = b.send_report('Test suite', 'Test fake credentials')
    assert ret is True


def test_fake_credentials_bad_repo():
    b = get_backend_bad_repo()
    b.get_user_credentials = get_fake_user_credentials
    ret = b.send_report('Test suite', 'Test fake credentials')
    assert ret is False


def test_get_credentials_from_qsettings():
    qsettings = qt.QtCore.QSettings('TestCrashCredentials')
    qsettings.clear()
    api.set_qsettings(qsettings)
    b = get_backend()
    username, remember = b._get_credentials_from_qsettings()
    assert username == ''
    assert remember is False

    qsettings.setValue('github/username', 'toto')
    qsettings.setValue('github/remember_credentials', '1')

    username, remember = b._get_credentials_from_qsettings()
    assert username == 'toto'
    assert remember is True


def test_store_user_credentials():
    qsettings = qt.QtCore.QSettings('TestCrashCredentials')
    qsettings.clear()
    api.set_qsettings(qsettings)
    b = get_backend()
    b._store_credentials('user', 'toto', False)
    username, remember = b._get_credentials_from_qsettings()
    assert username == 'user'
    assert remember is False
